from numpy import *
from matplotlib import pyplot 
from poisson import poisson

# SPLU is a sparse matrix LU factorization
from scipy.sparse.linalg import splu
# speye generates a sparse identity matrix
from scipy.sparse import eye as speye

# Task Overview:
# - Read through the code.  Make sure you understand the tasks. 
#
# - Implement the tasks required to run a study with backward Euler 
#
# - Carry out an error study with Backward Euler.  That is,   
#   --> Plot the error as you refine in h and ht.
#   --> How should h and ht decrease, relative to each other, in order to get 
#       balanced error?  Remember Euler is O(h) and the spatial discretization
#       is O(h^2)
#       --> This will tell you how to choose Num_time_steps and Num_space_pts
#           and observe O(h^2) error
#
# - Implement forward Euler
#   --> To run forward Euler, you'll have to comment in the correct time 
#       stepping scheme down below and comment in the Num_time_steps and 
#       Num_space_pts choices for forward Euler
#   --> If I told you that the  max eigenvalue (lambda) of A was roughly 8.0/h^2, 
#       Could you pre-compute the largest allowed time-step size ht?
#       
#       Derive a condition like   
#           ht < alpha h**2
#       where you have to find alpha.  Remember, the stability bound for forward
#       Euler that 
#           | 1  - ht lambda | < 1
#
#   --> Generating an error plot with forward Euler to show the error convergence.
#
# - If bored...you could change the exact solution and forcing, so that you have
#   a nonzero boundary term.


def euler_backward(A, u, ht, f):
    '''
    Carry out backward Euler's method for one time step
    '''
    Ainv = splu(#TASK: fill in the correct matrix to invert for for Euler's method) 
    return Ainv.solve(#TASK: fill in the correct right-hand-side for Euler's method)


def euler_forward(A, u, ht, f):
    '''
    Carry out forward Euler's method for one time step
    '''
    return u # + TASK: Fill in with correct values for forward Euler 


##
# Problem Definition
#
# u_t = u_xx + u_yy + f,    on the unit box [0,1] x [0,1] and t in a user-defined interval
#
# with an exact solution of
# u(t,x,y) = sin(t)*sin(pi*x)*sin(pi*y)
#
# which in turn, implies a forcing term of f, where
# f(t,x,y) = cos(t)*sin(pi*x)*sin(pi*y) + 2*pi**2*sin(t)*sin(pi*x)*sin(pi*y) 
#
# u(t=0,x,y) = 0            zero initial condition
#
# g(t,x,y) = u(t,x,y) = 0   for x,y on the boundary
#



# Declare the problem
def uexact(t,x,y):
    # Exact solution
    return sin(pi*t)*sin(pi*x)*sin(pi*y)

def f(t,x,y):
    # Forcing term
    # This should equal u_t - u_xx - u_yy
    return #Task: fill in 

##
# Comment in the number of time-steps and points in space to loop over 
#
# Backward Euler (time step, mesh size) combos that show O(h^2) error
Num_time_steps = array([#Task:  fill in time-step values that show O(h^2) error 
Num_space_pts = array([8, 16, 32, 64])
#
# Forward Euler stable (time step, mesh size) combos that show O(h^2) error
#Num_time_steps = array([#Task:  fill time-step values that are stable and show O(h^2) error 
#Num_space_pts = array([4, 8, 16, 32])

# Loop over various numbers of time points (nt) and spatial grid sizes (n)
enorm = zeros((len(Num_time_steps),))
counter = 0
for (nt, n) in zip(Num_time_steps, Num_space_pts):


    # Declare time domain
    t0 = 0.0
    T = 1.5
    ht = (T - t0)/float(nt-1)

    # Declare spatial domain
    h = 1.0 / (n-1.0)
    # Remember, we assume a zero boundary condition, which simplifies things a bit here
    # Thus, we only want a spatial grid from [h, 2h, ..., 1-h] x [h, 2h, ..., 1-h]
    pts = # Task: fill in with correct definition of pts
    X,Y = meshgrid(pts, pts)
    X = X.reshape(-1,)
    Y = Y.reshape(-1,)


    # Declare spatial discretization matrix
    # Task (pencil and paper): What dimension should A be?  
    #                          Remember the spatial grid is from [h, 2h, ..., 1-h] x [h, 2h, ..., 1-h]
    # Task: finish the below call to poisson() to generate A 
    A = poisson((sizex, sizey, format='csc')
    A = #Task: scale A by 1/h^2 


    # Declare initial condition
    #   This initial condition obeys an all 0 boundary condition, meaning the
    #   solution is clamped to 0 at the boundary.
    u0 = uexact(0, X, Y) 


    # Declare storage
    u = zeros((nt,A.shape[0]))
    ue = zeros((nt,A.shape[0]))
    u[0,:] = u0
    ue[0,:] = u0

    # Run time-stepping
    for i in range(1,nt):
        ue[i,:] = # Task: evaluate uexact to fill in this value 
        
        # Option 1: Backward Euler
        u[i,:] = euler_backward# Task: add call to backward euler 
                               #  --> Be careful which time value "t" you pass in to f(t,u)

        # Option 2: Forward Euler 
        #u[i,:] = euler_forward# Task: add call to forward euler 
                               #  --> Be careful which time value "t" you pass in to f(t,u)
        

    # Compute L2-norm of the error at final time
    e = (u - ue).reshape(-1,)
    enorm[counter] = # Task: compute the L2 norm over space-time here
    print("Nt, N, Error is:  " + str(nt) + ",  " + str(n) + ",  " + str(enorm[counter]))
    counter = counter+1

    # Task: Toggle the solution plotting to True, visualize the solution,
    #       understand that it is a composition of sine functions 
    if False:
        pyplot.figure(1)
        pyplot.imshow(u[0,:].reshape(n-2,n-2), origin='lower', extent=(0, 1, 0, 1))
        pyplot.colorbar()
        pyplot.xlabel('X')
        pyplot.ylabel('Y')
        pyplot.title("Initial Condition")
        
        pyplot.figure(2)
        pyplot.imshow(u[int(nt/2),:].reshape(n-2,n-2))
        pyplot.colorbar()
        pyplot.xlabel('X')
        pyplot.ylabel('Y')
        pyplot.title("Solution at time T/2")
        
        pyplot.figure(3)
        pyplot.imshow(u[-1,:].reshape(n-2,n-2))
        pyplot.colorbar()
        pyplot.xlabel('X')
        pyplot.ylabel('Y')
        pyplot.title("Solution at final time")
       
        pyplot.figure(4)
        pyplot.imshow(uexact(T,X,Y).reshape(n-2,n-2))
        pyplot.colorbar()
        pyplot.xlabel('X')
        pyplot.ylabel('Y')
        pyplot.title("Exact Solution at final time")
        
        pyplot.show()

##
# Plot the error, is it decreasing as expected, i.e., quadratically?
pyplot.figure(5)
pyplot.loglog(1.0/Num_space_pts, enorm, 'k') 
# Task: Add a reference quadratic curve to see if your error is decreasing quadratically
# Task: Add axis labels and a legend
pyplot.show()

