from numpy import *
from mpi4py import MPI

###
# Our goal is to solve a system 
#          A x = 0
# where A is 1D Poisson matrix with stencil
#         [1 -2 1]
#
# We don't build a CSR matrix here and the right-hand-side
# of the linear system is 0.  We make such a simple problem
# so that its easy to practice with MPI coding.  If you want
# you can extend the code to handle a nonzero right-hand-side. 
#
# Implement the below tasks, and make sure you read and 
# understand all the code.
#
###


##
# Initial MPI Commands
comm = MPI.COMM_WORLD           # Retrieve basic communicator and initialize MPI
rank = comm.Get_rank()          # Find out this processor's 
                                # rank.  If you ran 
                                #  $ mpirun -np k  python jacobi_oneD_MPI.py
                                # Then, there are k total ranks
num_ranks = comm.size           # Returns "k", the total num ranks
print("My rank is %d out of %d total processors"%(rank,num_ranks))


##
# Set problem parameters
n_global = 2*3*4*5              # Global matrix size [ divides evenly by 1, 2, 3, 4, 5, ... ]
n_local = n_global / num_ranks  # How many rows of the matrix does each proc own
n_local = int(n_local)
x_local = random.rand(n_local,) # Each proc starts with a random initial guess

# Task (pencil and paper): Determine what kind of boundary condition this sets 
#                          for the left-most and right-most processes 
x_left = array([0.0])           # Data to receive from left neighbor
x_right = array([0.0])          # Data to receive from right neighbor

maxiter = 10                    # Maximum allowed Jacobi iterations

for k in range(maxiter):
    
    # Jacobi requires duplicate storage for the solution
    x_local_new = zeros_like(x_local)

    # Send to left neighbor (if not rank 0)
    # --> Pay attention to the data-type
    if rank != 0:
        comm.Send([x_local[0:1], MPI.DOUBLE], dest=rank-1, tag=77)

    # Task: Send to right neighbor (if not last rank)
    if rank != (num_ranks-1):
        comm.Send([x_local[-1:], MPI.DOUBLE], dest=rank+1, tag=77)

    # Task: Receive from right neighbor (if not last rank)
    if rank != (num_ranks-1):
        comm.Recv([x_right[0:1], MPI.DOUBLE], source=rank+1, tag=77)

    # Task: Receive from left neighbor (if not rank 0)
    if rank != 0:
        comm.Recv([x_left[0:1], MPI.DOUBLE], source=rank-1, tag=77)
    
    # Compute the residual norm squared over each processor
    # Task: Understand why resid_global, and resid_local must be arrays (and not scalars)
    resid_global = array([0.0])
    resid_local = array([0.0])
    resid_local[0] += (-x_left[0] + 2*x_local[0] - x_local[1])**2
    resid_local[0] += (-x_local[-2] + 2*x_local[-1] - x_right[0])**2
    for i in range(1, n_local-1):
        resid_local[0] += (-x_local[i-1] + 2*x_local[i] - x_local[i+1])**2
    
    # Sum all local residual norms squared, and compute global residual norm
    #   Remember the interface to Allreduce: 
    #       Allreduce(local_value, global_reduced_value, reduction_op) 
    #   Note that our reduction_op is MPI.SUM
    # Task: complete the below Allreduce 
    comm.Allreduce(resid_local, resid_global, op=MPI.SUM) 
    resid_global[0] = sqrt(resid_global[0])
    
    # Task: Change this print statement so that only rank 0 prints
    if rank == 0:
        print("  Iteration " + str(k) + ", || r_k || = " + str(resid_global[0]))
    
    # Task: Insert a halting tolerance.  Should only rank 0 check this, or not?
    if resid_global[0] < 0.1:
        break

    # Task: Update first and last entries of x_local
    x_local_new[0] = (x_left[0] + x_local[1])/2.0
    x_local_new[-1] = (x_right[0] + x_local[-2])/2.0

    # Task: Update all interior points
    for i in range(1, n_local-1):
        x_local_new[i] = (x_local[i-1] + x_local[i+1])/2.0

    # Task: Understand why we overwrite x_local here.
    # Store new solution
    x_local = x_local_new
    

