from numpy import *
from matplotlib import pyplot

##
# Homework 5 Preliminary
# Practice implementing Euler, RK2, and RK4
##

def euler(f, y0, n, t0, T):
    '''
    Carry out Euler's method on [t0, T]
    for y' = f(t,y), nsteps n, and initial 
    condition y0.
    '''

    y = zeros((n,))
    y[0] = y0
    h = float(T - t0) / (n-1.0)

    for i in range(1, n):
        y[i] = y[i-1] + h*f(y[i-1], (i-1)*h)

    return y

def RK2(f, y0, n, t0, T):
    '''
    Carry out RK2 on [t0, T]
    for y' = f(t,y), nsteps n, and initial 
    condition y0.
    '''

    y = zeros((n,))
    y[0] = y0
    h = float(T - t0) / (n-1.0)

    for i in range(1, n):
        
        k1 = f(y[i-1], (i-1)*h)
        k2 = f(y[i-1] + h*k1, i*h)
        y[i] = y[i-1] + (h/2.0) * (k1 + k2)

    return y

def RK4(f, y0, n, t0, T):
    '''
    Carry out RK4 on [t0, T]
    for y' = f(t,y), nsteps n, and initial 
    condition y0.
    '''

    y = zeros((n,))
    y[0] = y0
    h = float(T - t0) / (n-1.0)

    for i in range(1, n):
        
        k1 = f(y[i-1], (i-1)*h)
        k2 = f(y[i-1] + (h/2.0)*k1, (i-1)*h + (h/2.0) )
        k3 = f(y[i-1] + (h/2.0)*k2, (i-1)*h + (h/2.0) )
        k4 = f(y[i-1] + h*k3, i*h)

        y[i] = y[i-1] + (h/6.0)*(k1 + 2*k2 + 2*k3 + k4) 

    return y

##
# Define your forcing function f(t,y) and your exact solution for checking error
##

def f(y, t):
    return t*y + t**3

def yexact(t):
    return 3*exp(t**2.0 / 2.0) - t**2 - 2.0


##
# Set up your problem domain
##

t0 = 0.0
T = 1.0
y0 = 1.0
n = 5

##
# Carry out your first numerical ODE experiments, solve the above problem from
# time t0 until time T, using n=5 time-steps.
yE = euler(f, y0, n, t0, T)
yRK2 = RK2(f, y0, n, t0, T)
yRK4 = RK4(f, y0, n, t0, T)


##
# Plotting the results for the above experiment.  
# The plotting is taken care of for you, don't need to change anything
##
pyplot.figure(1)
pts = linspace(t0, T, n)
manypts = linspace(t0, T, 2000)
pyplot.plot(pts, yE, linewidth='2') 
pyplot.plot(pts, yRK2, linewidth='4') 
pyplot.plot(pts, yRK4, linewidth='2') 
pyplot.plot(manypts, yexact(manypts), linewidth='2') 
pyplot.xlabel(r'$X$', fontsize='large')
pyplot.ylabel(r'$Y$', fontsize='large')
#pyplot.xticks() 
pyplot.legend(['Euler', 'RK2', 'RK4', 'Exact'], fontsize='large') 
pyplot.savefig('lec20_compare_E_RK2_RK4.png', dpi=500, format='png', bbox_inches='tight', pad_inches=0.0,)



##
# Next, we do a convergence study for 10 different problem sizes
##

EulerError = zeros((10,))
RK2Error = zeros((10,))
RK4Error = zeros((10,))

# These are the problem sizes that we will loop over
ns = 10*arange(1,100,10)

# Begin loop over the problem sizes
for i,n in enumerate(ns):
    
    h = float(T - t0) / (n-1.0)
    yE = euler(f, y0, n, t0, T)
    yRK2 = RK2(f, y0, n, t0, T)
    yRK4 = RK4(f, y0, n, t0, T)
    yExact = yexact(linspace(t0, T, n))

    EulerError[i] = sqrt(h*sum((yE - yExact)**2))
    RK2Error[i] = sqrt(h*sum((yRK2 - yExact)**2))
    RK4Error[i] = sqrt(h*sum((yRK4 - yExact)**2))
    

pyplot.figure(2)
pyplot.loglog(1.0/ns, EulerError, 'k') 
pyplot.loglog(1.0/ns, RK2Error, 'r') 
pyplot.loglog(1.0/ns, RK4Error, 'm') 
pyplot.loglog(1.0/ns, 1.0/ns, '--k') 
pyplot.loglog(1.0/ns, (1.0/ns)**2, '--r') 
pyplot.loglog(1.0/ns, (1.0/ns)**4, '--m') 
pyplot.xlabel(r'$h$', fontsize='large')
pyplot.ylabel('Error', fontsize='large')
#pyplot.xticks() 
pyplot.legend(['Euler', 'RK2', 'RK4', r'$h$', r'$h^2$', r'$h^4$'], fontsize='large') 

pyplot.savefig('lec20_compare_error_E_RK2_RK4.png', dpi=500, format='png', bbox_inches='tight', pad_inches=0.0,)


##
# These are reference plots for generating the stability regions for RK1, RK2, and RK4
if False:

    pyplot.figure(3)
    [x,y] = meshgrid( linspace(-4,4,1000), linspace(-4, 4, 1000))
    z = x + 1.0j*y
    pyplot.contour(x, y, abs(1 - z + 0.5*z**2 - (1/6.)*z**3 + (1/24.)*z**4), array([1.0]) )
    pyplot.xlabel(r'$h \lambda_R$', fontsize='large')
    pyplot.ylabel(r'$h \lambda_I$', fontsize='large')
    #pyplot.xticks() 
    pyplot.title('RK4 Stability Region', fontsize='large')
    pyplot.axis('equal')
    pyplot.grid()
    pyplot.savefig('lec20_RK4_contour.png', dpi=500, format='png', bbox_inches='tight', pad_inches=0.0,)

    pyplot.figure(4)
    [x,y] = meshgrid( linspace(-4,4,1000), linspace(-4, 4, 1000))
    z = x + 1.0j*y
    pyplot.contour(x, y, abs(1 - z + 0.5*z**2), array([1.0]) )
    pyplot.xlabel(r'$h \lambda_R$', fontsize='large')
    pyplot.ylabel(r'$h \lambda_I$', fontsize='large')
    #pyplot.xticks() 
    pyplot.title('RK2 Stability Region', fontsize='large')
    pyplot.axis('equal')
    pyplot.grid()
    pyplot.savefig('lec20_RK2_contour.png', dpi=500, format='png', bbox_inches='tight', pad_inches=0.0,)

    pyplot.figure(5)
    [x,y] = meshgrid( linspace(-4,4,1000), linspace(-4, 4, 1000))
    z = x + 1.0j*y
    pyplot.contour(x, y, abs(1 - z), array([1.0]) )
    pyplot.xlabel(r'$h \lambda_R$', fontsize='large')
    pyplot.ylabel(r'$h \lambda_I$', fontsize='large')
    #pyplot.xticks() 
    pyplot.title('RK1 Stability Region', fontsize='large')
    pyplot.axis('equal')
    pyplot.grid()
    pyplot.savefig('lec20_RK1_contour.png', dpi=500, format='png', bbox_inches='tight', pad_inches=0.0,)

